/*******************************************************************************

    nubd.c
    Copyright 2019 NetApp, Inc.

This file is derived from the Linux kernel rdma.c file
of the nvme driver.  rdma.c served as a model for the transport
layer, but the majority of the implementation has been rewritten
for our shared memory transport layer.

Change Log (for open source compliance):
    (2019-10-15) Added NUBD manager device and commented our calls to IB device driver.
    (2019-10-28) Updated controller struct and renamed for nubd.  Added code to create the NUBD controller device.
    (2019-11-14) Removed several IB-specific transport items and switched to a pseudo reg space.
    (2019-11-21) Updated a few of the mq_ops and IOCTLs
    (2019-11-25) Updated NUBD device connect sequence.

*******************************************************************************/

/*
 * NVMe over Fabrics RDMA host code.
 * Copyright (c) 2015-2016 HGST, a Western Digital Company.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 */
#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <rdma/mr_pool.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/atomic.h>
#include <linux/blk-mq.h>
#include <linux/blk-mq-rdma.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/miscdevice.h>
#include <linux/mutex.h>
#include <linux/scatterlist.h>
#include <linux/nvme.h>
#include <asm/unaligned.h>

#include <rdma/ib_verbs.h>
#include <rdma/rdma_cm.h>
#include <linux/nvme-rdma.h>

#include "uapi/linux/nubd.h"
#include "nvme.h"
#include "fabrics.h"


#define NVME_RDMA_CONNECT_TIMEOUT_MS    3000        /* 3 second */

#define NVME_RDMA_MAX_SEGMENTS      256

#define NVME_RDMA_MAX_INLINE_SEGMENTS   4

struct nvme_rdma_qe {
    struct ib_cqe   cqe;
    void            *data;
    u64             dma;
};

struct nvme_nubd_queue;
/* TODO ESL2C-244: Review and adjust contents of the request
 *          when working on IO through the coupling driver.
 */
struct nvme_rdma_request {
    struct nvme_request req;
    struct ib_mr        *mr;
    struct nvme_rdma_qe sqe;
    union nvme_result   result;
    __le16              status;
    refcount_t          ref;
    struct ib_sge       sge[1 + NVME_RDMA_MAX_INLINE_SEGMENTS];
    u32                 num_sge;
    int                 nents;
    struct ib_reg_wr    reg_wr;
    struct ib_cqe       reg_cqe;
    struct nvme_nubd_queue  *queue;
    struct sg_table     sg_table;
    struct scatterlist  first_sgl[];
};

struct io_region_entry {
    struct  io_region_entry* next;
    void*   v_addr;
    u64     p_addr;
    size_t  len;
};

/* Use atomic access functions when setting flags. */
enum nvme_nubd_queue_flags {
    NVME_NUBD_Q_ALLOCATED   = 0,
    NVME_NUBD_Q_LIVE        = 1,
};

struct nvme_nubd_queue {
    struct device *device;
    struct nvme_nubd_ctrl   *ctrl;
    struct workqueue_struct *io_workqueue;
    struct work_struct cq_update_work;
    char irqname[24];

    spinlock_t sq_lock;
    struct nvme_command *sq_cmds;
    spinlock_t cq_lock ____cacheline_aligned_in_smp;
    volatile struct nvme_completion *cqes;
    struct blk_mq_tags **tags;
    dma_addr_t sq_dma_addr;
    dma_addr_t cq_dma_addr;
    u32 __iomem *sq_db;
    u32 __iomem *cq_db;
    u16 queue_size;
    u16 sq_tail;
    u16 cq_head;
    u16 qid;
    u8  cq_phase;
    unsigned long flags;
};

typedef enum {
    NUBD_CTRL_STATE_OFF     = 0,
    NUBD_CTRL_STATE_INIT    = 1,
    NUBD_CTRL_STATE_STARTED = 2
} nubd_ctrl_state;

struct nvme_nubd_ctrl {
    /* read only in the hot path */
    struct nvme_nubd_queue  **queues;

    /* NUBD specific stuff we add */
    int                 instance;
    nubd_ctrl_state     state;
    void __iomem*       cmb_pbase;
    void*               cmb_vbase;
    size_t              cmb_unused_offset;
    size_t              cmb_size;
    void __iomem*       reg_base;
    u32                 aq_size;
    u32                 ioq_size;
    u32                 max_num_ioq;
    u32*                db_buf;
    u32                 db_stride;
    char                nqn[NVMF_NQN_FIELD_LEN];
    char                name[12];
    struct kref         ref;
    struct miscdevice   miscdev;
    struct io_region_entry* io_region_list;

    /* other member variables */
    struct blk_mq_tag_set   tag_set;
    struct work_struct  err_work;

    struct nvme_rdma_qe async_event_sqe;

    struct delayed_work reconnect_work;

    struct list_head    list;

    struct blk_mq_tag_set   admin_tag_set;

    struct sockaddr_storage addr;
    struct sockaddr_storage src_addr;

    struct nvme_ctrl    ctrl;
    bool                use_inline_data;
};

static void cq_update_workfn(struct work_struct *work);

static inline struct nvme_nubd_ctrl *to_nubd_ctrl(struct nvme_ctrl *ctrl)
{
    return container_of(ctrl, struct nvme_nubd_ctrl, ctrl);
}

/* This is redundant currently with the nubd_ctrls array, but we kept it around
 * as a list would be more flexible that the array if we ever support multiple
 * controllers.  As it's not in the IO path it doesn't have any significant impact.
 */
static LIST_HEAD(nvme_nubd_ctrl_list);
static DEFINE_MUTEX(nvme_nubd_ctrl_mutex);

/*
 * Disabling this option makes small I/O goes faster, but is fundamentally
 * unsafe.  With it turned off we will have to register a global rkey that
 * allows read and write access to all physical memory.
 */
static bool register_always = true;
module_param(register_always, bool, 0444);
MODULE_PARM_DESC(register_always,
     "Use memory registration even for contiguous memory regions");

/*
 * We should be able to support up to 26, but keep this as a smaller value
 * as we really only expect to have one controller for the foreseeable future.
 */
static unsigned char max_ctrls = 4;
module_param(max_ctrls, byte, 0644);
MODULE_PARM_DESC(max_ctrls, "Maximum number of nubd controller devices");

static int nubd_major;
module_param(nubd_major, int, 0444);
MODULE_PARM_DESC(nubd_major, "Major number to be used for block devices");

static int clnt_completion = 1;
module_param(clnt_completion, int, 0644);
MODULE_PARM_DESC(clnt_completion, "Completions handled in client context (performance tunable)");


static DEFINE_IDA(nubd_ctrl_ida);

/* This macro translates a dma address in the CMB to a kernel virtual address */
#define KV_CMB_ADDR(dev, addr)  (void*) (dev->cmb_vbase + ((uintptr_t)addr - (uintptr_t)dev->cmb_pbase))
#define NUB_KDEV(ndev)  ndev->miscdev.this_device

#define PAGE_ROUND_UP(addr)     round_up((uintptr_t)addr, PAGE_SIZE)

/*
 * The application should be polling, so technically writing the doorbell
 * value should not be needed.  We do it here to stay congruent to the hardware
 * implementation, and as it may provide some debug information if things
 * go wrong.
 * Memory barriers ensure that the updates to the queue have made it out
 * to memory before the doorbell update does.
 */
#define SQ_DB_NOTIFY(tail, db_addr)     do { wmb(); writel(tail, db_addr); } while (0)

static struct miscdevice nubd_mgr;

/* Tracking array for nubd controller device instances. */
static struct nvme_nubd_ctrl** nubd_ctrls;

static const struct blk_mq_ops nvme_rdma_mq_ops;
static const struct blk_mq_ops nvme_rdma_admin_mq_ops;

/* XXX: really should move to a generic header sooner or later.. */
static inline void put_unaligned_le24(u32 val, u8 *p)
{
    *p++ = val;
    *p++ = val >> 8;
    *p++ = val >> 16;
}

static int nubd_wait_ready(struct nvme_nubd_ctrl *ctrl, bool enabled)
{
    unsigned long timeout;
    u32 bit = enabled ? NVME_CSTS_RDY : 0;
    u64 cap = ctrl->ctrl.cap;

    /* If no timeout value, set to 3 seconds, which should be more than sufficient.
     * cap register timeout value is in 500 msec intervals.
     */
    timeout = NVME_CAP_TIMEOUT(cap) ? NVME_CAP_TIMEOUT(cap) : 6;
    timeout = ((timeout * HZ) / 2) + jiffies;

    while ((readl(ctrl->reg_base + NVME_REG_CSTS) & NVME_CSTS_RDY) != bit) {
        msleep(100);
        if (fatal_signal_pending(current))
            return -EINTR;
        if (time_after(jiffies, timeout)) {
            dev_err(NUB_KDEV(ctrl),
                "Device not ready; aborting %s\n", enabled ?
                        "initialization" : "reset");
            return -ENODEV;
        }
    }

    return 0;
}

static int nubd_enable_controller(struct nvme_nubd_ctrl *ctrl)
{
    int error = 0;
    u64 cc;

    cc = readl(ctrl->reg_base + NVME_REG_CC) | NVME_CC_ENABLE;
    writel(cc, ctrl->reg_base + NVME_REG_CC);
    error = nubd_wait_ready(ctrl, true);
    if (error)
        return error;

    /* User space driver is expected to go off registers above, but to keep
     * things consistent we also write it through the fabric mechanism.
     */
    error = nvme_enable_ctrl(&ctrl->ctrl, ctrl->ctrl.cap);
    return error;
}

static int nubd_disable_controller(struct nvme_nubd_ctrl *ctrl)
{
    int error = 0;
    u64 cc;

    /* User space driver is expected to go off registers, but to keep
     * things consistent we also write it through the fabric mechanism.
     */
    error = nvme_disable_ctrl(&ctrl->ctrl, ctrl->ctrl.cap);
    if (error)
        return error;

    cc = readl(ctrl->reg_base + NVME_REG_CC) & ~NVME_CC_ENABLE;
    writel(cc, ctrl->reg_base + NVME_REG_CC);
    error = nubd_wait_ready(ctrl, false);
    return error;
}

static void nvme_rdma_exit_request(struct blk_mq_tag_set *set,
        struct request *rq, unsigned int hctx_idx)
{
/* TODO ESL2C-244: figure out req mgmt for basic IO. */
#if 0
    struct nvme_nubd_ctrl *ctrl = set->driver_data;
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);
    int queue_idx = (set == &ctrl->tag_set) ? hctx_idx + 1 : 0;
    struct nvme_nubd_queue *queue = ctrl->queues[queue_idx];

    nvme_rdma_free_qe(NUB_KDEV(ctrl), &req->sqe, sizeof(struct nvme_command),
            DMA_TO_DEVICE);
#endif
}

static int nvme_rdma_init_request(struct blk_mq_tag_set *set,
        struct request *rq, unsigned int hctx_idx,
        unsigned int numa_node)
{
/* TODO ESL2C-244: figure out req mgmt for basic IO. */
#if 0
    struct nvme_nubd_ctrl *ctrl = set->driver_data;
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);
    int queue_idx = (set == &ctrl->tag_set) ? hctx_idx + 1 : 0;
    struct nvme_nubd_queue *queue = ctrl->queues[queue_idx];
    int ret = 0;

    nvme_req(rq)->ctrl = &ctrl->ctrl;
    ret = nvme_rdma_alloc_qe(ibdev, &req->sqe, sizeof(struct nvme_command),
            DMA_TO_DEVICE);
    if (ret)
        return ret;

    req->queue = queue;

#endif
    return 0;
}

static int nvme_nubd_init_hctx(struct blk_mq_hw_ctx *hctx, void *data,
        unsigned int hctx_idx)
{
    struct nvme_nubd_ctrl *ctrl = data;
    struct nvme_nubd_queue *queue = ctrl->queues[hctx_idx + 1];

    BUG_ON(hctx_idx >= ctrl->ctrl.queue_count);

    hctx->driver_data = queue;
    return 0;
}

static int nvme_nubd_init_admin_hctx(struct blk_mq_hw_ctx *hctx, void *data,
        unsigned int hctx_idx)
{
    struct nvme_nubd_ctrl *ctrl = data;
    struct nvme_nubd_queue *queue = ctrl->queues[0];

    BUG_ON(hctx_idx != 0);

    hctx->driver_data = queue;
    return 0;
}

static int nvme_nubd_alloc_queue(struct nvme_nubd_ctrl *ctrl,
        int idx, size_t queue_size)
{
    struct nvme_nubd_queue *queue;
    dma_addr_t  region_end;
    struct workqueue_struct *wq;
    int result;

    queue = kzalloc(sizeof(*queue), GFP_KERNEL);
    if (!queue)
        return -ENOMEM;

    snprintf(queue->irqname, sizeof(queue->irqname), "nvme%dq%d",
            ctrl->instance, idx);

    /*
     * The intention is that the user space part of the driver will signal us via the
     * workqueue whenever there is work to do.  We might also consider some poll
     * cycle every few seconds to check the queue for pending work in case the
     * message somehow gets missed.
     */
    wq = create_singlethread_workqueue(queue->irqname);
    if (!wq)
    {
        result = -ENOMEM;
        goto free_queue;
    }

    queue->io_workqueue = wq;
    INIT_WORK( &queue->cq_update_work, cq_update_workfn );

    queue->ctrl = ctrl;
    queue->device = NUB_KDEV(ctrl);
    queue->qid = idx;
    queue->queue_size = queue_size;

    /* Make sure the device stays around as long as the queue */
    kref_get(&ctrl->ref);

    queue->sq_db = ctrl->db_buf + (idx * 2) * ctrl->db_stride;
    queue->cq_db = ctrl->db_buf + (idx * 2 + 1) * ctrl->db_stride;
    dev_dbg(NUB_KDEV(ctrl), "Q IDX %d - SQ DB:%lx, CQ DB: %lx.\n", idx, (uintptr_t)queue->sq_db, (uintptr_t)queue->cq_db);

    queue->sq_dma_addr = PAGE_ROUND_UP(ctrl->cmb_pbase + ctrl->cmb_unused_offset);
    region_end = queue->sq_dma_addr + (queue_size * sizeof(struct nvme_command));
    queue->cq_dma_addr = PAGE_ROUND_UP(region_end);
    region_end = queue->cq_dma_addr + (queue_size * sizeof(struct nvme_completion));
    dev_dbg(NUB_KDEV(ctrl), "  - SQ base:%llx, CQ base: %llx.\n", queue->sq_dma_addr, queue->cq_dma_addr);
    if ( ctrl->cmb_size < (region_end - (uintptr_t)ctrl->cmb_pbase))
    {
        /* Not enough memory in CMB... something was not valid about the configuration.
         * Clear out the struct and log an error.
         */
        dev_err(NUB_KDEV(ctrl),
            "Not enough CMB left for queue creation:cmb_size(0x%lx), cmb_unused_offset(0x%lx),\n"
            "                                       q index(%d), q size(0x%lx)\n",
            ctrl->cmb_size, ctrl->cmb_unused_offset, idx, queue_size);
        result = -EINVAL;
        goto free_wq;
    }
    ctrl->cmb_unused_offset = region_end - (uintptr_t)ctrl->cmb_pbase;

    queue->sq_cmds = KV_CMB_ADDR(ctrl, queue->sq_dma_addr);
    queue->cqes = KV_CMB_ADDR(ctrl, queue->cq_dma_addr);

    spin_lock_init(&queue->sq_lock);
    spin_lock_init(&queue->cq_lock);

    /* queue->tags is set up during the connection phase. */
    queue->sq_tail = 0;
    queue->cq_head = 0;
    queue->cq_phase = 1;
    set_bit(NVME_NUBD_Q_ALLOCATED, &queue->flags);

    ctrl->queues[idx] = queue;
    return 0;

free_wq:
    destroy_workqueue(queue->io_workqueue);
free_queue:
    kfree(queue);
    return result;
}

static void nubd_free_dev(struct kref *kref)
{
    struct nvme_nubd_ctrl *dev = container_of(kref, struct nvme_nubd_ctrl, ref);

    /* Everything else should be freed during controller shutdown. */
    kfree(dev);
}

static void nvme_nubd_stop_queue(struct nvme_nubd_queue *queue)
{
    if (!test_and_clear_bit(NVME_NUBD_Q_LIVE, &queue->flags))
        return;

/* For now just update the flag value.  If we find that we need
 * a transport-specific behavior to stop queue operation,
 * it should go here.
 */
}

static void nvme_nubd_free_queue(struct nvme_nubd_queue *queue)
{
    if (!test_and_clear_bit(NVME_NUBD_Q_ALLOCATED, &queue->flags))
        return;

    destroy_workqueue(queue->io_workqueue);
    kfree(queue);
    kref_put(&queue->ctrl->ref, nubd_free_dev);
}

static void nvme_nubd_free_io_queues(struct nvme_nubd_ctrl *ctrl)
{
    int i;

    for (i = 1; i < ctrl->ctrl.queue_count; i++)
    {
        nvme_nubd_free_queue(ctrl->queues[i]);
        ctrl->queues[i] = 0;
    }
}

static void nvme_nubd_stop_io_queues(struct nvme_nubd_ctrl *ctrl)
{
    int i;

    for (i = 1; i < ctrl->ctrl.queue_count; i++)
        nvme_nubd_stop_queue(ctrl->queues[i]);
}

static int nvme_nubd_start_queue(struct nvme_nubd_ctrl *ctrl, int idx)
{
    int ret;

    if (idx)
        ret = nvmf_connect_io_queue(&ctrl->ctrl, idx);
    else
        ret = nvmf_connect_admin_queue(&ctrl->ctrl);

    if (!ret)
    {
        set_bit(NVME_NUBD_Q_LIVE, &ctrl->queues[idx]->flags);
        dev_dbg(NUB_KDEV(ctrl), "Connected to queue %d.\n", idx);
    }
    else
        dev_info(NUB_KDEV(ctrl),
            "failed to connect queue: %d ret=%d\n", idx, ret);

    return ret;
}

static int nvme_nubd_start_io_queues(struct nvme_nubd_ctrl *ctrl)
{
    int i, ret = 0;

    for (i = 1; i < ctrl->ctrl.queue_count; i++) {
        ret = nvme_nubd_start_queue(ctrl, i);
        if (ret)
            goto out_stop_queues;
    }

    return 0;

out_stop_queues:
    for (i--; i >= 1; i--)
        nvme_nubd_stop_queue(ctrl->queues[i]);
    return ret;
}

static int nvme_nubd_alloc_queues(struct nvme_nubd_ctrl *ctrl)
{
    unsigned int nr_io_queues;
    int ret;
    int i = 0;

    /* IO queues is the user space requested number of IO queues if
     * it is valid (>2), but otherwise will use the number of CPU cores.
     */
    nr_io_queues = (ctrl->max_num_ioq > 1) ? ctrl->max_num_ioq : num_online_cpus();

    /* list of NVME queue pair structures (plus 1 for admin queue) */
    ctrl->queues = kcalloc(nr_io_queues + 1, sizeof(*ctrl->queues),
                                GFP_KERNEL);
    if (!ctrl->queues)
        return -ENOMEM;

    ctrl->db_buf = ctrl->reg_base + NVME_REG_DBS;

    /* Admin Queue */
    dev_dbg(NUB_KDEV(ctrl), "creating Admin Queue.\n");
    ctrl->aq_size = NVME_AQ_DEPTH;
    ret = nvme_nubd_alloc_queue(ctrl, 0, ctrl->aq_size);
    if (ret)
        goto free_queues;

    /* IO Queues */
    dev_dbg(NUB_KDEV(ctrl), "creating %d I/O queues.\n", nr_io_queues);
    for (i = 1; i <= nr_io_queues; i++) {
        ret = nvme_nubd_alloc_queue(ctrl, i, ctrl->ioq_size);
        if (ret)
            goto free_queues;
    }

    ctrl->ctrl.queue_count = nr_io_queues + 1;
    return 0;

free_queues:
    for (i--; i >= 0; i--)
    {
        nvme_nubd_free_queue(ctrl->queues[i]);
        ctrl->queues[i] = 0;
    }
    kfree(ctrl->queues);
    ctrl->queues = 0;
    ctrl->cmb_unused_offset = 0;
    return ret;
}

static void nvme_nubd_free_tagset(struct nvme_ctrl *nctrl,
        struct blk_mq_tag_set *set)
{
    blk_mq_free_tag_set(set);
}

static struct blk_mq_tag_set *nvme_nubd_alloc_tagset(struct nvme_ctrl *nctrl,
        bool admin)
{
    struct nvme_nubd_ctrl *ctrl = to_nubd_ctrl(nctrl);
    struct blk_mq_tag_set *set;
    int ret;

    if (admin) {
        set = &ctrl->admin_tag_set;
        memset(set, 0, sizeof(*set));
        set->ops = &nvme_rdma_admin_mq_ops;
        set->queue_depth = NVME_AQ_MQ_TAG_DEPTH;
        set->reserved_tags = 2; /* connect + keep-alive */
        set->numa_node = NUMA_NO_NODE;
/* TODO ESL2C-244: Figure out what private data to  attach to IO requests from the mq. */
        set->cmd_size = sizeof(struct nvme_rdma_request) +
            SG_CHUNK_SIZE * sizeof(struct scatterlist);
        set->driver_data = ctrl;
        set->nr_hw_queues = 1;
        set->timeout = ADMIN_TIMEOUT;
        set->flags = BLK_MQ_F_NO_SCHED;
    } else {
        set = &ctrl->tag_set;
        memset(set, 0, sizeof(*set));
        set->ops = &nvme_rdma_mq_ops;
        set->queue_depth = nctrl->sqsize + 1;
        set->reserved_tags = 1; /* fabric connect */
        set->numa_node = NUMA_NO_NODE;
        set->flags = BLK_MQ_F_SHOULD_MERGE;
        set->cmd_size = sizeof(struct nvme_rdma_request) +
            SG_CHUNK_SIZE * sizeof(struct scatterlist);
        set->driver_data = ctrl;
        set->nr_hw_queues = nctrl->queue_count - 1;
        set->timeout = NVME_IO_TIMEOUT;
    }

    ret = blk_mq_alloc_tag_set(set);
    if (ret)
        goto out;

    return set;

out:
    return ERR_PTR(ret);
}

static void nvme_nubd_destroy_admin_queue(struct nvme_nubd_ctrl *ctrl,
        bool remove)
{
    if (remove) {
        blk_cleanup_queue(ctrl->ctrl.admin_q);
        nvme_nubd_free_tagset(&ctrl->ctrl, ctrl->ctrl.admin_tagset);
    }
    if (ctrl->async_event_sqe.data) {
        ctrl->async_event_sqe.data = NULL;
    }
    nvme_nubd_free_queue(ctrl->queues[0]);
    ctrl->queues[0] = 0;
}

static int nvme_nubd_configure_admin_queue(struct nvme_nubd_ctrl *ctrl,
        bool new)
{
    int error;

    if (new) {
        ctrl->ctrl.admin_tagset = nvme_nubd_alloc_tagset(&ctrl->ctrl, true);
        if (IS_ERR(ctrl->ctrl.admin_tagset)) {
            error = PTR_ERR(ctrl->ctrl.admin_tagset);
            goto out_free_async_qe;
        }

        ctrl->ctrl.admin_q = blk_mq_init_queue(&ctrl->admin_tag_set);
        if (IS_ERR(ctrl->ctrl.admin_q)) {
            error = PTR_ERR(ctrl->ctrl.admin_q);
            goto out_free_tagset;
        }
    }

    error = nvme_nubd_start_queue(ctrl, 0);
    if (error)
        goto out_cleanup_queue;

    ctrl->ctrl.sqsize =
        min_t(int, NVME_CAP_MQES(ctrl->ctrl.cap), ctrl->ctrl.sqsize);

    error = nubd_enable_controller(ctrl);
    if (error)
        goto out_stop_queue;

    /* ESL2C-244: At most 1 page.. check pci driver to see if we might
     * need to set other fields in the nvme_ctrl struct.
     */
    ctrl->ctrl.max_hw_sectors = 4096;

    error = nvme_init_identify(&ctrl->ctrl);
    if (error)
        goto out_disable_ctrl;

    return 0;

out_disable_ctrl:
    nubd_disable_controller(ctrl);
out_stop_queue:
    nvme_nubd_stop_queue(ctrl->queues[0]);
out_cleanup_queue:
    if (new)
        blk_cleanup_queue(ctrl->ctrl.admin_q);
out_free_tagset:
    if (new)
        nvme_nubd_free_tagset(&ctrl->ctrl, ctrl->ctrl.admin_tagset);
out_free_async_qe:
    ctrl->async_event_sqe.data = NULL;
    return error;
}

static void nvme_nubd_destroy_io_queues(struct nvme_nubd_ctrl *ctrl,
        bool remove)
{
    if (remove) {
        blk_cleanup_queue(ctrl->ctrl.connect_q);
        nvme_nubd_free_tagset(&ctrl->ctrl, ctrl->ctrl.tagset);
    }
    nvme_nubd_free_io_queues(ctrl);
}

static int nvme_nubd_configure_io_queues(struct nvme_nubd_ctrl *ctrl, bool new)
{
    int ret;

    if (new) {
        ctrl->ctrl.tagset = nvme_nubd_alloc_tagset(&ctrl->ctrl, false);
        if (IS_ERR(ctrl->ctrl.tagset)) {
            return(PTR_ERR(ctrl->ctrl.tagset));
        }

        ctrl->ctrl.connect_q = blk_mq_init_queue(&ctrl->tag_set);
        if (IS_ERR(ctrl->ctrl.connect_q)) {
            ret = PTR_ERR(ctrl->ctrl.connect_q);
            goto out_free_tag_set;
        }
    } else {
        blk_mq_update_nr_hw_queues(&ctrl->tag_set,
            ctrl->ctrl.queue_count - 1);
    }

    ret = nvme_nubd_start_io_queues(ctrl);
    if (ret)
        goto out_cleanup_connect_q;

    return 0;

out_cleanup_connect_q:
    if (new)
        blk_cleanup_queue(ctrl->ctrl.connect_q);
out_free_tag_set:
    if (new)
        nvme_nubd_free_tagset(&ctrl->ctrl, ctrl->ctrl.tagset);
    return ret;
}

static void nvme_nubd_teardown_admin_queue(struct nvme_nubd_ctrl *ctrl,
        bool remove)
{
    blk_mq_quiesce_queue(ctrl->ctrl.admin_q);
    nvme_nubd_stop_queue(ctrl->queues[0]);
    blk_mq_tagset_busy_iter(&ctrl->admin_tag_set, nvme_cancel_request,
            &ctrl->ctrl);
    blk_mq_unquiesce_queue(ctrl->ctrl.admin_q);
    nvme_nubd_destroy_admin_queue(ctrl, remove);
}

static void nvme_nubd_teardown_io_queues(struct nvme_nubd_ctrl *ctrl,
        bool remove)
{
    if (ctrl->ctrl.queue_count > 1) {
        nvme_stop_queues(&ctrl->ctrl);
        nvme_nubd_stop_io_queues(ctrl);
        blk_mq_tagset_busy_iter(&ctrl->tag_set, nvme_cancel_request,
                &ctrl->ctrl);
        if (remove)
            nvme_start_queues(&ctrl->ctrl);
        nvme_nubd_destroy_io_queues(ctrl, remove);
    }
}

static void nvme_nubd_stop_ctrl(struct nvme_ctrl *nctrl)
{
    struct nvme_nubd_ctrl *ctrl = to_nubd_ctrl(nctrl);

    cancel_work_sync(&ctrl->err_work);
    cancel_delayed_work_sync(&ctrl->reconnect_work);
}

static void nvme_nubd_free_ctrl(struct nvme_ctrl *nctrl)
{
    struct nvme_nubd_ctrl *ctrl = to_nubd_ctrl(nctrl);

    if (list_empty(&ctrl->list))
        goto free_ctrl;

    mutex_lock(&nvme_nubd_ctrl_mutex);
    list_del(&ctrl->list);
    mutex_unlock(&nvme_nubd_ctrl_mutex);

    nvmf_free_options(nctrl->opts);
free_ctrl:
    kfree(ctrl->queues);
    kfree(ctrl);
}

static void nvme_rdma_reconnect_or_remove(struct nvme_nubd_ctrl *ctrl)
{
    /* If we are resetting/deleting then do nothing */
    if (ctrl->ctrl.state != NVME_CTRL_CONNECTING) {
        WARN_ON_ONCE(ctrl->ctrl.state == NVME_CTRL_NEW ||
            ctrl->ctrl.state == NVME_CTRL_LIVE);
        return;
    }

    if (nvmf_should_reconnect(&ctrl->ctrl)) {
        dev_info(ctrl->ctrl.device, "Reconnecting in %d seconds...\n",
            ctrl->ctrl.opts->reconnect_delay);
        queue_delayed_work(nvme_wq, &ctrl->reconnect_work,
                ctrl->ctrl.opts->reconnect_delay * HZ);
    } else {
        nvme_delete_ctrl(&ctrl->ctrl);
    }
}

static int nvme_nubd_setup_ctrl(struct nvme_nubd_ctrl *ctrl, bool new)
{
    int ret;
    bool changed;

    ret = nvme_nubd_configure_admin_queue(ctrl, new);
    if (ret)
        return ret;

    ret = -EINVAL;
    if (ctrl->ctrl.icdoff) {
        dev_err(ctrl->ctrl.device, "icdoff is not supported!\n");
        goto uninit_adminq;
    }

    if (!(ctrl->ctrl.sgls & (1 << 2))) {
        dev_err(ctrl->ctrl.device,
            "Mandatory keyed sgls are not supported!\n");
        goto uninit_adminq;
    }

    if (ctrl->ctrl.opts->queue_size > ctrl->ctrl.sqsize + 1) {
        dev_warn(ctrl->ctrl.device,
            "queue_size %zu > ctrl sqsize %u, clamping down\n",
            ctrl->ctrl.opts->queue_size, ctrl->ctrl.sqsize + 1);
    }

    if (ctrl->ctrl.sqsize + 1 > ctrl->ctrl.maxcmd) {
        dev_warn(ctrl->ctrl.device,
            "sqsize %u > ctrl maxcmd %u, clamping down\n",
            ctrl->ctrl.sqsize + 1, ctrl->ctrl.maxcmd);
        ctrl->ctrl.sqsize = ctrl->ctrl.maxcmd - 1;
    }

    if (ctrl->ctrl.queue_count > 1) {
        ret = nvme_nubd_configure_io_queues(ctrl, new);
        if (ret)
            goto uninit_adminq;
    }

    changed = nvme_change_ctrl_state(&ctrl->ctrl, NVME_CTRL_LIVE);
    if (!changed) {
        /* state change failure is ok if we're in DELETING state */
        WARN_ON_ONCE(ctrl->ctrl.state != NVME_CTRL_DELETING);
        ret = -EINVAL;
        goto uninit_ioq;
    }

    nvme_start_ctrl(&ctrl->ctrl);
    return 0;

uninit_ioq:
    nvme_nubd_stop_io_queues(ctrl);
    blk_cleanup_queue(ctrl->ctrl.connect_q);
    nvme_nubd_free_tagset(&ctrl->ctrl, ctrl->ctrl.tagset);
uninit_adminq:
    nubd_disable_controller(ctrl);
    nvme_nubd_stop_queue(ctrl->queues[0]);
    blk_cleanup_queue(ctrl->ctrl.admin_q);
    nvme_nubd_free_tagset(&ctrl->ctrl, ctrl->ctrl.admin_tagset);
    ctrl->async_event_sqe.data = NULL;

    return ret;
}

static void nvme_rdma_reconnect_ctrl_work(struct work_struct *work)
{
    struct nvme_nubd_ctrl *ctrl = container_of(to_delayed_work(work),
            struct nvme_nubd_ctrl, reconnect_work);

    ++ctrl->ctrl.nr_reconnects;

    if (nvme_nubd_setup_ctrl(ctrl, false))
        goto requeue;

    dev_info(ctrl->ctrl.device, "Successfully reconnected (%d attempts)\n",
            ctrl->ctrl.nr_reconnects);

    ctrl->ctrl.nr_reconnects = 0;

    return;

requeue:
    dev_info(ctrl->ctrl.device, "Failed reconnect attempt %d\n",
            ctrl->ctrl.nr_reconnects);
    nvme_rdma_reconnect_or_remove(ctrl);
}

static void nvme_rdma_error_recovery_work(struct work_struct *work)
{
    struct nvme_nubd_ctrl *ctrl = container_of(work,
            struct nvme_nubd_ctrl, err_work);

    nvme_stop_keep_alive(&ctrl->ctrl);
    nvme_nubd_teardown_io_queues(ctrl, false);
    nvme_start_queues(&ctrl->ctrl);
    nvme_nubd_teardown_admin_queue(ctrl, false);

    if (!nvme_change_ctrl_state(&ctrl->ctrl, NVME_CTRL_CONNECTING)) {
        /* state change failure is ok if we're in DELETING state */
        WARN_ON_ONCE(ctrl->ctrl.state != NVME_CTRL_DELETING);
        return;
    }

    nvme_rdma_reconnect_or_remove(ctrl);
}

static void nvme_rdma_error_recovery(struct nvme_nubd_ctrl *ctrl)
{
    if (!nvme_change_ctrl_state(&ctrl->ctrl, NVME_CTRL_RESETTING))
        return;

    queue_work(nvme_wq, &ctrl->err_work);
}

static void nvme_rdma_wr_error(struct ib_cq *cq, struct ib_wc *wc,
        const char *op)
{
    struct nvme_nubd_queue *queue = cq->cq_context;
    struct nvme_nubd_ctrl *ctrl = queue->ctrl;

    if (ctrl->ctrl.state == NVME_CTRL_LIVE)
        dev_info(NUB_KDEV(ctrl),
                 "%s for CQE 0x%p failed with status (%d)\n",
                 op, wc->wr_cqe, wc->status);
    nvme_rdma_error_recovery(ctrl);
}

static void nvme_rdma_inv_rkey_done(struct ib_cq *cq, struct ib_wc *wc)
{
    struct nvme_rdma_request *req =
        container_of(wc->wr_cqe, struct nvme_rdma_request, reg_cqe);
    struct request *rq = blk_mq_rq_from_pdu(req);

    if (unlikely(wc->status != IB_WC_SUCCESS)) {
        nvme_rdma_wr_error(cq, wc, "LOCAL_INV");
        return;
    }

    if (refcount_dec_and_test(&req->ref))
        nvme_end_request(rq, req->status, req->result);

}

static int nvme_rdma_inv_rkey(struct nvme_nubd_queue *queue,
        struct nvme_rdma_request *req)
{
/* TODO ESL2C-293: Implement lkey/rkey or other security options. */
    return 0;
}

static void nvme_rdma_unmap_data(struct nvme_nubd_queue *queue,
        struct request *rq)
{
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);

    if (!blk_rq_payload_bytes(rq))
        return;

    if (req->mr) {
        /* ib_mr_pool_put(queue->qp, &queue->qp->rdma_mrs, req->mr); */
        req->mr = NULL;
    }

    nvme_cleanup_cmd(rq);
    sg_free_table_chained(&req->sg_table, true);
}

static int nvme_rdma_set_sg_null(struct nvme_command *c)
{
    struct nvme_keyed_sgl_desc *sg = &c->common.dptr.ksgl;

    sg->addr = 0;
    put_unaligned_le24(0, sg->length);
    put_unaligned_le32(0, sg->key);
    sg->type = NVME_KEY_SGL_FMT_DATA_DESC << 4;
    return 0;
}

static int nvme_rdma_map_data(struct nvme_nubd_queue *queue,
        struct request *rq, struct nvme_command *c)
{
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);
    int count, ret;

    req->num_sge = 1;
    refcount_set(&req->ref, 2); /* send and recv completions */

    c->common.flags |= NVME_CMD_SGL_METABUF;

    if (!blk_rq_payload_bytes(rq))
        return nvme_rdma_set_sg_null(c);

    req->sg_table.sgl = req->first_sgl;
    ret = sg_alloc_table_chained(&req->sg_table,
            blk_rq_nr_phys_segments(rq), req->sg_table.sgl);
    if (ret)
        return -ENOMEM;

    req->nents = blk_rq_map_sg(rq->q, rq, req->sg_table.sgl);

/* TODO ESL2C-244: Do we need to do any mapping here? Or even need this function? */

    count = req->nents;
    if (unlikely(count <= 0)) {
        ret = -EIO;
        goto out_free_table;
    }

    return 0;

out_free_table:
    sg_free_table_chained(&req->sg_table, true);
    return ret;
}

static void nvme_rdma_send_done(struct ib_cq *cq, struct ib_wc *wc)
{
    struct nvme_rdma_qe *qe =
        container_of(wc->wr_cqe, struct nvme_rdma_qe, cqe);
    struct nvme_rdma_request *req =
        container_of(qe, struct nvme_rdma_request, sqe);
    struct request *rq = blk_mq_rq_from_pdu(req);

    if (unlikely(wc->status != IB_WC_SUCCESS)) {
        nvme_rdma_wr_error(cq, wc, "SEND");
        return;
    }

    if (refcount_dec_and_test(&req->ref))
        nvme_end_request(rq, req->status, req->result);
}

static struct blk_mq_tags *nvme_nubd_tagset(struct nvme_nubd_queue *queue)
{
    u32 queue_idx = queue->qid;

    if (queue_idx == 0)
        return queue->ctrl->admin_tag_set.tags[queue_idx];
    return queue->ctrl->tag_set.tags[queue_idx - 1];
}

static int nvme_submit_cmd(struct nvme_nubd_queue* q, struct nvme_command* cmd)
{
    spin_lock(&q->sq_lock);
    memcpy(&q->sq_cmds[q->sq_tail], cmd, sizeof(*cmd));

    if (++q->sq_tail == q->queue_size)
        q->sq_tail = 0;

    SQ_DB_NOTIFY(q->sq_tail, q->sq_db);

    spin_unlock(&q->sq_lock);

    return 0;
}

static void nvme_nubd_submit_async_event(struct nvme_ctrl *arg)
{
    struct nvme_nubd_ctrl *ctrl = to_nubd_ctrl(arg);
    struct nvme_nubd_queue *admin_q = ctrl->queues[0];
    struct nvme_command c;

    memset(&c, 0, sizeof(c));
    c.common.opcode = nvme_admin_async_event;
    c.common.command_id = NVME_AQ_BLK_MQ_DEPTH;
    nvme_submit_cmd(admin_q, &c);
}

static int nvme_rdma_process_nvme_rsp(struct nvme_nubd_queue *queue,
        struct nvme_completion *cqe, struct ib_wc *wc, int tag)
{
    struct request *rq;
    struct nvme_rdma_request *req;
    int ret = 0;

    rq = blk_mq_tag_to_rq(nvme_nubd_tagset(queue), cqe->command_id);
    if (!rq) {
        dev_err(queue->ctrl->ctrl.device,
            "tag 0x%x on qid %d not found\n",
            cqe->command_id, queue->qid);
        nvme_rdma_error_recovery(queue->ctrl);
        return ret;
    }
    req = blk_mq_rq_to_pdu(rq);

    req->status = cqe->status;
    req->result = cqe->result;

    if (wc->wc_flags & IB_WC_WITH_INVALIDATE) {
        if (unlikely(wc->ex.invalidate_rkey != req->mr->rkey)) {
            dev_err(queue->ctrl->ctrl.device,
                "Bogus remote invalidation for rkey %#x\n",
                req->mr->rkey);
            nvme_rdma_error_recovery(queue->ctrl);
        }
    } else if (req->mr) {
        ret = nvme_rdma_inv_rkey(queue, req);
        if (unlikely(ret < 0)) {
            dev_err(queue->ctrl->ctrl.device,
                "Queueing INV WR for rkey %#x failed (%d)\n",
                req->mr->rkey, ret);
            nvme_rdma_error_recovery(queue->ctrl);
        }
        /* the local invalidation completion will end the request */
        return 0;
    }

    if (refcount_dec_and_test(&req->ref)) {
        if (rq->tag == tag)
            ret = 1;
        nvme_end_request(rq, req->status, req->result);
    }

    return ret;
}

static enum blk_eh_timer_return
nvme_rdma_timeout(struct request *rq, bool reserved)
{
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);
    struct nvme_nubd_queue *queue = req->queue;
    struct nvme_nubd_ctrl *ctrl = queue->ctrl;

    dev_warn(ctrl->ctrl.device, "I/O %d QID %d timeout\n",
         rq->tag, queue->qid);

    if (ctrl->ctrl.state != NVME_CTRL_LIVE) {
        /*
         * Teardown immediately if controller times out while starting
         * or we are already started error recovery. all outstanding
         * requests are completed on shutdown, so we return BLK_EH_DONE.
         */
        flush_work(&ctrl->err_work);
        nvme_nubd_teardown_io_queues(ctrl, false);
        nvme_nubd_teardown_admin_queue(ctrl, false);
        return BLK_EH_DONE;
    }

    dev_warn(ctrl->ctrl.device, "starting error recovery\n");
    nvme_rdma_error_recovery(ctrl);

    return BLK_EH_RESET_TIMER;
}

static blk_status_t nvme_rdma_queue_rq(struct blk_mq_hw_ctx *hctx,
        const struct blk_mq_queue_data *bd)
{
    struct nvme_ns *ns = hctx->queue->queuedata;
    struct nvme_nubd_queue *queue = hctx->driver_data;
    struct request *rq = bd->rq;
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);
    struct nvme_command c;
    bool queue_ready = test_bit(NVME_NUBD_Q_LIVE, &queue->flags);
    blk_status_t ret;
    int err;

    WARN_ON_ONCE(rq->tag < 0);

    if (!nvmf_check_ready(&queue->ctrl->ctrl, rq, queue_ready))
        return nvmf_fail_nonready_command(&queue->ctrl->ctrl, rq);

    ret = nvme_setup_cmd(ns, rq, &c);
    if (ret)
        return ret;

    blk_mq_start_request(rq);
    if (c.connect.opcode==nvme_fabrics_command && c.connect.fctype==nvme_fabrics_type_connect)
    {
        /* We're already connected if we are this far, so send back good status. */
        blk_mq_end_request(rq, BLK_STS_OK);
    }

    err = nvme_rdma_map_data(queue, rq, &c);
    if (unlikely(err < 0)) {
        dev_err(queue->ctrl->ctrl.device,
                 "Failed to map data (%d)\n", err);
        nvme_cleanup_cmd(rq);
        goto err;
    }

    err = nvme_submit_cmd(queue, &c);
    if (unlikely(err)) {
        dev_err(queue->ctrl->ctrl.device,
                 "Command submission failed (qid %d, error %d)\n", queue->qid, err);
        goto err;
    }

    return BLK_STS_OK;
err:
    if (err == -ENOMEM || err == -EAGAIN)
        return BLK_STS_RESOURCE;
    return BLK_STS_IOERR;
}

/* We read the CQE phase first to check if the rest of the entry is valid */
static inline bool nubd_cqe_pending(struct nvme_nubd_queue *q)
{
    return (le16_to_cpu(q->cqes[q->cq_head].status) & 1) ==
            q->cq_phase;
}

static inline void nubd_handle_cqe(struct nvme_nubd_queue *q, u16 idx)
{
    volatile struct nvme_completion *cqe = &q->cqes[idx];
    struct request *req;

    if (unlikely(cqe->command_id >= q->queue_size)) {
        dev_warn(q->ctrl->ctrl.device,
            "invalid id %d completed on queue %d\n",
            cqe->command_id, le16_to_cpu(cqe->sq_id));
        return;
    }

    /*
     * AEN requests are special as they don't time out and can
     * survive any kind of queue freeze and often don't respond to
     * aborts.  We don't even bother to allocate a struct request
     * for them but rather special case them here.
     */
    if (unlikely(q->qid == 0 &&
            cqe->command_id >= NVME_AQ_BLK_MQ_DEPTH)) {
        nvme_complete_async_event(&q->ctrl->ctrl,
                cqe->status, &cqe->result);
        return;
    }

    req = blk_mq_tag_to_rq(*q->tags, cqe->command_id);
    nvme_end_request(req, cqe->status, cqe->result);
}

static void nubd_complete_cqes(struct nvme_nubd_queue *q, u16 start, u16 end)
{
    while (start != end) {
        nubd_handle_cqe(q, start);
        if (++start == q->queue_size)
            start = 0;
    }
}

/* Spinlocks should be taken before calling this function. */
static inline bool nubd_process_cq(struct nvme_nubd_queue *q, u16 *start,
        u16 *end, int tag)
{
    bool found = false;

    *start = q->cq_head;
    while (!found && nubd_cqe_pending(q))
    {
        if (q->cqes[q->cq_head].command_id == tag)
            found = true;

        if (q->cq_head == q->queue_size - 1)
        {
            q->cq_head = 0;
            q->cq_phase = !q->cq_phase;
        }
        else
        {
            q->cq_head++;
        }
    }
    *end = q->cq_head;

    if (*start != *end)
        writel(q->cq_head, q->cq_db);

    return found;
}

static int __nubd_poll(struct nvme_nubd_queue* q, unsigned int tag)
{
    u16 start, end;
    bool found;

    if (!nubd_cqe_pending(q))
        return 0;

    spin_lock_irq(&q->cq_lock);
    found = nubd_process_cq(q, &start, &end, tag);
    spin_unlock_irq(&q->cq_lock);

    nubd_complete_cqes(q, start, end);
    return found;
}

static int nvme_nubd_poll(struct blk_mq_hw_ctx *hctx, unsigned int tag)
{
    int found = 0;
    struct nvme_nubd_queue *queue = hctx->driver_data;

    found = __nubd_poll(queue, tag);
    return found;
}

static void cq_update_workfn(struct work_struct *work)
{
    struct nvme_nubd_queue *cq = container_of(work, struct nvme_nubd_queue, cq_update_work);
    __nubd_poll(cq, -1);
}

static void nvme_rdma_complete_rq(struct request *rq)
{
    struct nvme_rdma_request *req = blk_mq_rq_to_pdu(rq);

    nvme_rdma_unmap_data(req->queue, rq);
    nvme_complete_rq(rq);
}

/*
 * This is actually default behavior for leaving .map_queues blank.
 * However it was included as we may want to adjust it in the future for
 * performance experiments, and this would give us a convenient place to do that.
 */
static int nvme_nubd_map_queues(struct blk_mq_tag_set *set)
{
    return blk_mq_map_queues(set);
}

static const struct blk_mq_ops nvme_rdma_mq_ops = {
    .queue_rq   = nvme_rdma_queue_rq,
    .complete   = nvme_rdma_complete_rq,
    .init_request   = nvme_rdma_init_request,
    .exit_request   = nvme_rdma_exit_request,
    .init_hctx  = nvme_nubd_init_hctx,
    .poll       = nvme_nubd_poll,
    .timeout    = nvme_rdma_timeout,
    .map_queues = nvme_nubd_map_queues,
};

static const struct blk_mq_ops nvme_rdma_admin_mq_ops = {
    .queue_rq   = nvme_rdma_queue_rq,
    .complete   = nvme_rdma_complete_rq,
    .init_request   = nvme_rdma_init_request,
    .exit_request   = nvme_rdma_exit_request,
    .init_hctx  = nvme_nubd_init_admin_hctx,
    .timeout    = nvme_rdma_timeout,
};

static void nvme_nubd_shutdown_ctrl(struct nvme_nubd_ctrl *ctrl, bool shutdown)
{
    struct io_region_entry* io_region;
    struct io_region_entry* tmp_region;

    nubd_disable_controller(ctrl);
    nvme_nubd_teardown_io_queues(ctrl, shutdown);
    if (shutdown)
    {
        for (tmp_region = ctrl->io_region_list; tmp_region; )
        {
            io_region = tmp_region;
            tmp_region = io_region->next;
            iounmap(io_region->v_addr);
            kfree(io_region);
        }
    }

    nvme_nubd_teardown_admin_queue(ctrl, shutdown);

    kfree(ctrl->queues);
    ctrl->queues = 0;
    ctrl->cmb_unused_offset = 0;
}

static void nvme_nubd_delete_ctrl(struct nvme_ctrl *ctrl)
{
    nvme_nubd_shutdown_ctrl(to_nubd_ctrl(ctrl), true);
}

static void nvme_rdma_reset_ctrl_work(struct work_struct *work)
{
    struct nvme_nubd_ctrl *ctrl =
        container_of(work, struct nvme_nubd_ctrl, ctrl.reset_work);

    nvme_stop_ctrl(&ctrl->ctrl);
    nvme_nubd_shutdown_ctrl(ctrl, false);

    if (!nvme_change_ctrl_state(&ctrl->ctrl, NVME_CTRL_CONNECTING)) {
        /* state change failure should never happen */
        WARN_ON_ONCE(1);
        return;
    }

    if (nvme_nubd_setup_ctrl(ctrl, false))
        goto out_fail;

    return;

out_fail:
    ++ctrl->ctrl.nr_reconnects;
    nvme_rdma_reconnect_or_remove(ctrl);
}

static const struct nvme_ctrl_ops nvme_nubd_ctrl_ops = {
    .name           = "nubd",
    .module         = THIS_MODULE,
    .flags          = NVME_F_FABRICS,
/* TODO ESL2C-316: do we need a wrapper to help access the reset/enable bits? */
    .reg_read32     = nvmf_reg_read32,
    .reg_read64     = nvmf_reg_read64,
    .reg_write32    = nvmf_reg_write32,
    .free_ctrl      = nvme_nubd_free_ctrl,
    .submit_async_event = nvme_nubd_submit_async_event,
    .delete_ctrl    = nvme_nubd_delete_ctrl,
    .get_address    = nvmf_get_address,
    .stop_ctrl      = nvme_nubd_stop_ctrl,
};

static int nubd_ctrl_open(struct inode *inode, struct file *f)
{
    struct nvme_nubd_ctrl *dev = container_of(f->private_data, struct nvme_nubd_ctrl,
                                miscdev);
    kref_get(&dev->ref);
    f->private_data = dev;
    return 0;
}

static int nubd_ctrl_release(struct inode *inode, struct file *f)
{
    struct nvme_nubd_ctrl *dev = f->private_data;
    kref_put(&dev->ref, nubd_free_dev);
    return 0;
}

static long nubd_ctrl_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
    struct nvme_nubd_ctrl* dev = f->private_data;
    struct nubd_io_mem_range new_iomem;
    struct io_region_entry* new_region;
    int qid;

    /* Only let Admin CQ tail updates through if the device is not started.
     * Admin CQ updates are necessary to complete device startup.
     */
    if((cmd!=NUBD_IOCTL_CQ_TAIL_UPDATE)
        && (dev->state != NUBD_CTRL_STATE_STARTED))
        return -ENOTTY;

/* TODO ESL2C-244: Implement NUBD controller IOCTLS */
    switch (cmd) {
    case NUBD_IOCTL_START_CTRL:
    case NUBD_IOCTL_RESCAN_DISKS:
    case NUBD_IOCTL_DISK_UPDATE:
    case NUBD_IOCTL_RD_COMPLETE:
    case NUBD_IOCTL_WR_COMPLETE:
        dev_warn(NUB_KDEV(dev), "Unimplemented ioctl request (0x%x)\n", cmd);
        return -ENOTTY;
    case NUBD_IOCTL_CQ_TAIL_UPDATE:
        qid = (int) arg;
        if ((qid < 0) || (qid > dev->ctrl.queue_count))
            return -EINVAL;
        /*
         * Running vdbench we can get a good boost in IOPs and a smaller but noticeable
         * one for bandwidth if we defer the completion queue processing to the client
         * context.  If the container needs a bit more CPU horsepower the clnt_completion
         * parameter can be cleared for the system to have the user space driver
         * incur that overhead instead.
         */
        if (clnt_completion)
        {
            queue_work( dev->queues[qid]->io_workqueue, &dev->queues[qid]->cq_update_work );
        }
        else
        {
            __nubd_poll(dev->queues[qid], -1);
        }
        return 0;
    case NUBD_IOCTL_ADD_IOMEM:
        if (copy_from_user( &new_iomem, (struct nubd_io_mem_range  __user*) arg,
                            sizeof(new_iomem)))
            return -EFAULT;

        /* would keeping the region list sorted help? */
        /* Do we need to check for overlapping/duplicate regions? */

        new_region = kzalloc(sizeof(struct io_region_entry), GFP_KERNEL);
        if (!new_region)
            return -ENOMEM;
        new_region->p_addr = new_iomem.p_addr;
        new_region->len = new_iomem.length;
        new_region->v_addr = ioremap_cache( new_iomem.p_addr,  new_iomem.length);
        new_region->next = dev->io_region_list;

        if ( !new_region->v_addr )
        {
            dev_err(NUB_KDEV(dev), "Failed mapping iomem region (0x%llx:0x%zx)\n",
                    new_iomem.p_addr, new_iomem.length);
            kfree(new_region);
            return -EINVAL;
        }

        dev->io_region_list = new_region;
        dev_dbg(NUB_KDEV(dev), "Mapped in region (0x%llx:0x%zx) at %p.\n",
            new_iomem.p_addr, new_iomem.length, new_region->v_addr);

        return 0;

    default:
        dev_warn(NUB_KDEV(dev), "Bad ioctl request (0x%x)\n", cmd);
        return -ENOTTY;
    }
}

static const struct file_operations nubd_ctrl_fops = {
    .owner      = THIS_MODULE,
    .open       = nubd_ctrl_open,
    .release    = nubd_ctrl_release,
    .unlocked_ioctl = nubd_ctrl_ioctl,
    .compat_ioctl   = nubd_ctrl_ioctl,
};

/* Initializes the device structure.
 *
 * dev: Pointer to zeroed-out nvme_nubd_ctrl structure.
 * usrconfig: pointer to struct from user with parameters of new device.
 *
 * This function is to set up the device structure, ready the admin queue for use
 * and have the memory spaces for the ioqueues figured out.
 *
 * We defer namespace discovery to another step to make sure the user space
 * part of the system is ready for it.
 */
static int nubd_ctrl_init(struct nvme_nubd_ctrl* dev,
        struct nubd_fabric_ctrl_config* usrConfig)
{
    int result = -ENOMEM;
    u32 max_io_queues;
    u64 cap_reg;
    u32 aqaVal;
    size_t len = strnlen(usrConfig->nqn, sizeof(usrConfig->nqn));

    pr_debug("Ctrl init start.\n");
    /* Sanity check given base values and cmb size */
    if ( !(usrConfig->cmb_base)
        || (usrConfig->cmb_size < 1024*1024)
        || ((uintptr_t)usrConfig->cmb_base & (PAGE_SIZE-1))
        || !(usrConfig->reg_base)
        || (usrConfig->reg_size <= NVME_REG_DBS) // At least enough to get to the DB regs
        || !len || (len == sizeof(usrConfig->nqn)))
    {
        print_hex_dump_bytes("usrConfig dump:", DUMP_PREFIX_NONE,  usrConfig, sizeof(*usrConfig));
        return -EINVAL;
    }

    /*
     * We might be able to reduce this with input, but for now
     * just allocate for all possible cpus.  Should be relatively small
     * amount of "wasted" memory.
     */
    max_io_queues = num_possible_cpus();

    /*
     *  Note: Current naming convention gives us 26 possible controller devices,
     *      nubd-a through nubd-z.  However we only use one at this point.
     *  We need to create the device prior to calling nvme_dev_start() so it
     *  can use the controller ioctls for adminQ update notification.
     */
    pr_debug("Adding Ctrl device.\n");
    snprintf(dev->name, sizeof(dev->name), "nubd-%c", 'a' + dev->instance );
    dev->miscdev.minor = MISC_DYNAMIC_MINOR;
    dev->miscdev.name = dev->name;
    dev->miscdev.parent = nubd_mgr.this_device;
    dev->miscdev.fops = &nubd_ctrl_fops;
    result = misc_register(&dev->miscdev);
    if (result)
        goto miscdev_error;

    /* Set up cmb for use. */
    dev->cmb_pbase = usrConfig->cmb_base;
    dev->cmb_unused_offset = 0;
    dev_dbg(NUB_KDEV(dev), "Mapping CMB (0x%lx) of size 0x%zx\n", (uintptr_t)usrConfig->cmb_base, usrConfig->cmb_size);
    dev->cmb_vbase = ioremap_cache((phys_addr_t)usrConfig->cmb_base, usrConfig->cmb_size);
    dev->cmb_size = usrConfig->cmb_size;

    /* Map in pseudo-reg space provided by user-space. */
    dev_dbg(NUB_KDEV(dev), "Mapping pseudo regs (0x%lx) of size 0x%zx\n", (uintptr_t)usrConfig->reg_base, usrConfig->reg_size);
    dev->reg_base = ioremap_cache((phys_addr_t)usrConfig->reg_base, usrConfig->reg_size);
    strncpy(dev->nqn, usrConfig->nqn, sizeof(dev->nqn));

    /* Copy off queue parameters for queue creation in the alloc routine and later. */
    dev->state = NUBD_CTRL_STATE_OFF;
    dev->max_num_ioq = usrConfig->num_io_queues;

    cap_reg = readq(dev->reg_base + NVME_REG_CAP);
    dev->ioq_size = NVME_CAP_MQES(cap_reg) + 1;
    if (dev->ioq_size < 2)
    {
        result = -EINVAL;
        goto unmap;
    }
    dev->ctrl.cap = cap_reg;
    dev->db_stride = 1 << (NVME_CAP_STRIDE(cap_reg));
    strncpy(dev->nqn, usrConfig->nqn, len);

    kref_init(&dev->ref);

    result = nvme_nubd_alloc_queues(dev);
    if (result)
    {
        dev_warn(NUB_KDEV(dev), "Queue allocation Failed! (result %d).\n", result);
        goto unmap;
    }

    dev_dbg(NUB_KDEV(dev), "Controller queues allocated.\n");
    /* Populate admin queue location info in struct to pass back to user space */
    aqaVal = dev->queues[0]->queue_size - 1;
    aqaVal |= aqaVal << 16;
    writel(aqaVal, dev->reg_base + NVME_REG_AQA);
    writeq(dev->queues[0]->sq_dma_addr, dev->reg_base + NVME_REG_ASQ);
    writeq(dev->queues[0]->cq_dma_addr, dev->reg_base + NVME_REG_ACQ);

    dev->io_region_list = NULL;
    dev->state = NUBD_CTRL_STATE_INIT;
    dev_dbg(NUB_KDEV(dev), "NUBD controller dev created.\n");
    return 0;

 unmap:
    iounmap(dev->reg_base);
    iounmap(dev->cmb_vbase);
    misc_deregister(&dev->miscdev);
 miscdev_error:
    return result;
}

/*
 * We only really support one controller now, but we tried to put in the
 * framework for being extensible if we ever found we needed it.
 * We only actually create the queue structures and SGL pools here.
 * The controller device is created later on the call to connect.
 */

static int nubd_mgr_create_ctrl(struct nubd_fabric_ctrl_config *ctrl_config)
{
    int result = 0;
    int new_dev_num;
    struct nvme_nubd_ctrl* new_ctrl;

    pr_info("NUBD Mgr - creating controller.\n");

    /* create char-misc device.  Add to list. */
    new_ctrl = kzalloc( sizeof(struct nvme_nubd_ctrl), GFP_KERNEL);
    if (!new_ctrl)
        return -ENOMEM;

    new_dev_num = ida_simple_get(&nubd_ctrl_ida, 0, max_ctrls, GFP_KERNEL);
    if (new_dev_num < 0)
    {
        kfree(new_ctrl);
        return new_dev_num;
    }

    new_ctrl->instance = new_dev_num;

    result = nubd_ctrl_init(new_ctrl, ctrl_config);

    pr_info("NUBD Mgr - controller init result: %d\n", result);
    if (result)
    {
        ida_simple_remove(&nubd_ctrl_ida, new_dev_num);
        kfree(new_ctrl);
    }
    else
    {
        ctrl_config->handle = new_dev_num;
        nubd_ctrls[new_dev_num] = new_ctrl;
    }

    ctrl_config->handle = new_dev_num;

    pr_debug("NUBD Mgr - End controller creation.\n");
    return result;
}

static int nubd_mgr_destroy_ctrl(int nubd_ctrl_num)
{
    struct nvme_nubd_ctrl *dev;

    /* Sanity check the given device handle. */
    if ( (nubd_ctrl_num >= max_ctrls)
        || (nubd_ctrl_num < 0)
        || !(nubd_ctrls[nubd_ctrl_num])
       )
    {
        return -EINVAL;
    }

    dev = nubd_ctrls[nubd_ctrl_num];

    nvme_nubd_delete_ctrl(&dev->ctrl);

    nubd_ctrls[nubd_ctrl_num] = 0;
    return 0;
}


static long nubd_mgr_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
    struct nubd_fabric_ctrl_config* new_ctrl_cfg;
    int version;
    int result = 0;

    pr_debug("NUBD mgr IOCTL 0x%x, arg 0x%lx,\n", cmd, arg);
    switch (cmd) {
    case NUBD_IOCTL_CREATE_DEV:
        /* A request for PCI controller creation is not supported
         * with this version of the driver.
         */
        return -ENOTTY;
    case NUBD_IOCTL_CREATE_FABRIC_DEV:
        /* allocating from the heap in case config structure gets big. */
        new_ctrl_cfg = kzalloc(sizeof(struct nubd_fabric_ctrl_config), GFP_KERNEL);
        if (!new_ctrl_cfg)
            return -ENOMEM;

        if (copy_from_user( new_ctrl_cfg, (struct nubd_fabric_ctrl_config __user*) arg,
                            sizeof(struct nubd_fabric_ctrl_config)))
        {
            result = -EFAULT;
            goto free_config_mem;
        }

        result = nubd_mgr_create_ctrl(new_ctrl_cfg);
        if (copy_to_user((struct nubd_fabric_ctrl_config __user*) arg, new_ctrl_cfg,
                            sizeof(struct nubd_fabric_ctrl_config)))
        {
            /* Unlikely the above will fail, but to be consistent we
             * should tear down the controller if we're returning error.
             */
            nubd_mgr_destroy_ctrl((int) arg);
            result = -EFAULT;
        }

free_config_mem:
        kfree(new_ctrl_cfg);
        return result;
    case NUBD_IOCTL_DESTROY_DEV:
        return nubd_mgr_destroy_ctrl((int) arg);
    case NUBD_IOCTL_GET_NUBD_REV:
        version = NUBD_FABRICS;
        if (copy_to_user((int*)arg, &version, sizeof(int)))
            return -EFAULT;
        else
            return 0;
    default:
        return -ENOTTY;
    }
}

static const struct file_operations nubd_mgr_fops = {
    .owner      = THIS_MODULE,
    .unlocked_ioctl = nubd_mgr_ioctl,
    .compat_ioctl   = nubd_mgr_ioctl
};

static struct nvme_ctrl *nvme_nubd_connect_ctrl(struct device *dev,
        struct nvmf_ctrl_options *opts)
{
    struct nvme_nubd_ctrl *ctrl;
    int ret;
    int i;
    bool changed;
    unsigned int nr_io_queues;

    for (i=0; i<max_ctrls; i++)
    {
        ctrl = nubd_ctrls[i];
        if (strncmp(ctrl->nqn, opts->subsysnqn, sizeof(ctrl->nqn)) == 0)
            break;
    }
    /* Check for invalid subsysnqn or if requested subsystem not found. */
    if (i == max_ctrls || !strnlen(opts->subsysnqn, sizeof(opts->subsysnqn)))
        return ERR_PTR(-ENODEV);

    /* We don't handle duplicate connections.  Return error if the
     * controller is already connected.
     */
    if (ctrl->state == NUBD_CTRL_STATE_STARTED)
    {
        return ERR_PTR(-EALREADY);
    }

    ctrl->ctrl.opts = opts;
    INIT_LIST_HEAD(&ctrl->list);

/* TODO ESL2C-244: Evaluate what we need to do here for
 *      controller device resets (if anything).
 */
#if 0
    INIT_LIST_HEAD(&dev->namespaces);
    dev->reset_workfn = nvme_reset_failed_dev;
    INIT_WORK(&dev->reset_work, nvme_reset_workfn);
#endif

    INIT_DELAYED_WORK(&ctrl->reconnect_work,
            nvme_rdma_reconnect_ctrl_work);
    INIT_WORK(&ctrl->err_work, nvme_rdma_error_recovery_work);
    INIT_WORK(&ctrl->ctrl.reset_work, nvme_rdma_reset_ctrl_work);

/* TODO ESL2C-244: Consider NVME_QUIRK_SIZE_STRIPE */
    ret = nvme_init_ctrl(&ctrl->ctrl, dev, &nvme_nubd_ctrl_ops,
                0 /* no quirks, we're perfect! */);
    if (ret)
        goto out_init_err;

    changed = nvme_change_ctrl_state(&ctrl->ctrl, NVME_CTRL_CONNECTING);
    WARN_ON_ONCE(!changed);

    /* Should be exactly one less, but as long as it's smaller
     * we should be OK.
     */
    if (opts->nr_io_queues >= ctrl->ctrl.queue_count)
    {
        ret = -EINVAL;
        goto out_uninit_ctrl;
    }

    /* This count possible leave some IOQs that were created unused, but
     * that may be OK if that was what is requested with this connection.
     */
    nr_io_queues = min(opts->nr_io_queues, ctrl->ctrl.queue_count-1);
    ret = nvme_set_queue_count(&ctrl->ctrl, &nr_io_queues);
    if (ret)
    {
        ret = -EIO;
        goto out_uninit_ctrl;
    }

    ctrl->ctrl.queue_count = nr_io_queues + 1; /* +1 for admin queue */
    ctrl->ctrl.kato = opts->kato;

    ret = nvme_nubd_setup_ctrl(ctrl, true);
    if (ret)
        goto out_uninit_ctrl;

    dev_info(ctrl->ctrl.device, "new ctrl: NQN \"%s\"\n",
        ctrl->ctrl.opts->subsysnqn);

    nvme_get_ctrl(&ctrl->ctrl);

    mutex_lock(&nvme_nubd_ctrl_mutex);
    list_add_tail(&ctrl->list, &nvme_nubd_ctrl_list);
    mutex_unlock(&nvme_nubd_ctrl_mutex);

    return &ctrl->ctrl;

out_uninit_ctrl:
    nvme_uninit_ctrl(&ctrl->ctrl);
    if (ret > 0)
        ret = -EIO;
out_init_err:
    return ERR_PTR(ret);
}

static struct nvmf_transport_ops nvme_nubd_transport = {
    .name       = "nubd",
    .module     = THIS_MODULE,
    .required_opts  = NVMF_OPT_NQN,
    .allowed_opts   = NVMF_OPT_KATO | NVMF_OPT_RECONNECT_DELAY |
              NVMF_OPT_CTRL_LOSS_TMO | NVMF_OPT_NR_IO_QUEUES,
    .create_ctrl    = nvme_nubd_connect_ctrl,
};

static int __init nvme_nubd_init_module(void)
{
    int ret;

    nubd_ctrls = kzalloc( max_ctrls * sizeof(struct nvme_nubd_ctrl*), GFP_KERNEL);
    if (!nubd_ctrls)
    {
        return -ENOMEM;
    }

    nubd_mgr.minor = MISC_DYNAMIC_MINOR;
    nubd_mgr.name = "nubd";
    nubd_mgr.fops = &nubd_mgr_fops;
    ret = misc_register (&nubd_mgr);
    if (ret)
        goto free_dev_array;

    ret = register_blkdev(nubd_major, "nubd");
    if (ret < 0)
    {
        goto unregister_mgr;
    }
    else if (ret > 0)
        nubd_major = ret;

    ret = nvmf_register_transport(&nvme_nubd_transport);
    if (ret)
        goto unregister_blkdev;

    return 0;

unregister_blkdev:
    unregister_blkdev(nubd_major, "nubd");
unregister_mgr:
    misc_deregister(&nubd_mgr);
free_dev_array:
    kfree(nubd_ctrls);
    return ret;
}

static void __exit nvme_nubd_cleanup_module(void)
{
    nvmf_unregister_transport(&nvme_nubd_transport);
    unregister_blkdev(nubd_major, "nubd");
    misc_deregister(&nubd_mgr);
    kfree(nubd_ctrls);
}

module_init(nvme_nubd_init_module);
module_exit(nvme_nubd_cleanup_module);

MODULE_LICENSE("GPL v2");
